import os
import numpy as np

class Fasta_segment():
    '''
    Efficient reads of a segments starting from a given offset
    from a FASTA file.

    ================================================================
    This class can be used ONLY for Fasta files where ALL sequence
    rows (excluding headers) are of the same length !!
    ================================================================

    In case of a FASTA with multiple sequences (multiple headers), the header name
    of the corresponding sequence must be provided. The header name is the string
    that proceeds the character ">".
    For example, the header name ofthe header:

    ">NC_000011.10 Homo sapiens chromosome 11, GRCh38.p13 Primary Assembly"

    is "NC_000011.10".

    In case of a FASTA with a single sequence (single header), the header name
    is not required.

    The class reads only the segment into memory and not the all FASTA sequence.
    This can be used, for example, to read a segment from a FASTA file
    that contains a chromosome of the human genome.
    '''
    def __init__(self, files_info=None) -> None:
        '''
        files_info (optional input) is a dictionary:
        For a FASTA with multiple headers, the key is a fasta file name and
        the header name as follows: "<fasta file name>:<header name>".
        For a FASTA with a single header the key is simply the file name (the header
        can also be provided, in which case the key will include it as well, however this is not required).
        The value is a dictionary that contains basic statistics of the fasta file.
        Keys of files_info[file] are:
            'init_loc': offset of the first NT in file (0 indicated the first NT)
            'l_size': number of NTs per line excluding \n,
            'line_size': number of NTs per line including \n.
            'file_size': total number of NTs in file, excluding \n (and excluding the header). This is
                         available only for FASTA files with a single header.

            For example: files_info = {'file1.fas': {'init_loc': 70, 'l_size': 80, 'line_size': 81, 'file_size': 240},
                               'file2.fas:NC_000011.10': {'init_loc': 70, 'l_size': 80, 'line_size': 81}}

        '''
        self.files_info = files_info if files_info else {}

    def __repr__(self) -> str:
        return f"Fasta_segment([files_info])"

    def __str__(self) -> str:
        return f"Efficient reads of segments from FASTA files.\n"

    def __len__(self) -> int:
        return len(self.files_info)

    def show_processed_files(self) -> list:
        return list(self.files_info.keys())

    def show_files_info(self) -> None:
        print(self.str_files_info())

    def str_files_info(self) -> str:
        s = ''
        for k in self.files_info.keys():
            s += f"{k}:\n"
            for ki, vi in self.files_info[k].items():
                s += f"\t{ki}: {vi}\n"

        return s

    def read_segment(self, file: str, offset: int, size: int) -> str:
        '''
        Reads a segment from a FASTA with a single sequence (single header).
        file - FASTA file name
        offset - 0 based offset of the start segment in file (from the first sequence NT).
                 For example, value of 1 indicates starting from the second NT in the file.
        size - number of NTs in a segment to read.
        '''
        if not os.path.exists(file):
            print(f"Error: file {file} does not exists !!")
            return ""

        with open(file, 'rt') as fp:
            if file in self.files_info:
                l_size = self.files_info[file]['l_size']
                line_size = self.files_info[file]['line_size']
                init_loc = self.files_info[file]['init_loc']
            else:
                init_loc, line_size, l_size = self.__compute_file_stats(fp)
                self.files_info[file] = {'init_loc': init_loc, 'l_size': l_size, 'line_size': line_size}

            offset_num_lines, offset_in_line = offset // l_size, np.mod(offset, l_size)
            # accounting for extra characters due to multiple \n that will be discarded
            add_for_newline = ((offset + size -1) // l_size) - offset_num_lines
            fp.seek(init_loc + offset_num_lines *line_size + offset_in_line)
            return fp.read(size + add_for_newline).replace('\n', '')

    def read_segment_endpoints(self, file: str, start_loc: int, end_loc: int):
        seq = Fasta_segment().read_segment(file, start_loc - 1, end_loc - start_loc + 1).upper()
        indices = list(range(start_loc, end_loc + 1))
        assert len(seq) == len(
            indices), f'reference data not compatible; {len(seq)}, {len(indices)}, start: {start_loc}, end: {end_loc}, {file}'
        return seq, indices

    def __compute_file_stats(self, fp) -> tuple:
        '''
        Computs the stats that are needed to read a segment from
        a FASTA file with a single sequence.
        '''
        # account for a header (if exists) in the first line of the file
        fp.seek(0, os.SEEK_SET)
        if fp.read(1) == ">":
            fp.readline()
            init_loc = fp.tell()
        else:
            init_loc = 0

        # init_loc is the location (0-based) of the first NT in file
        fp.seek(init_loc)
        fp.readline()  # advance handle to begining of second line
        line_size = fp.tell() - init_loc # number of NTs per line, including the \n
        return init_loc, line_size, line_size -1


    def total_num_chars(self, file: str) -> int:
        '''
        Returns the total number of characters (NTs) in a FASTA
        file (excluding the header and \n) without reading the all file.

        This function supports only Fasta with a single headr.
        Support for FASTAs with multiple headers is TBD.
        '''
        if file in self.files_info:
            return self.files_info[file]['file_size']
        else:
            return self.__compute_total_num_NTs(file)


    def __compute_total_num_NTs(self, file: str) -> int:
        '''
        Computes the total number of characters (NTs) in a FASTA
        file (excluding the header and \n) and updates self.files_info.
        '''
        with open(file, 'rt') as fp:
            init_loc, line_size, l_size = self.__compute_file_stats(fp)

            # check if last character is new line
            last_loc = fp.seek(0, os.SEEK_END)
            fp.seek(fp.tell() - 1, os.SEEK_SET)
            last_char = fp.read()

        delta = last_loc -init_loc
        q, r = divmod(delta, line_size)
        num_chars = delta - q # remove counts of new lines
        if q== 1 and r == 0 and (last_char != '\n'):
            num_chars += 1  # if only one line in file and last character is not newline add one

        self.files_info[file] = {'init_loc': init_loc,
                                 'l_size': l_size,
                                 'line_size': line_size,
                                 'file_size': num_chars}

        return num_chars

    def multiple_headers_read_segment(self, file: str, header_name: str, offset: int, size: int) -> str:
        '''
        Reades from a FASTA with multiple sequences (multiple headers).
        The offset and size here are relative to the sequence with header name header_name.

        For example, for header_name='name1', offset=3, size=4, the return sequence is of size 4 NTs, starting
        from offset 3 of the sequence that follows the header with header name header_name.
        '''
        init_loc = None
        token = file + ':' + header_name
        with open(file, 'rt') as fp:
            if token in self.files_info:
                l_size = self.files_info[token]['l_size']
                line_size = self.files_info[token]['line_size']
                init_loc = self.files_info[token]['init_loc']
            else:
                line = fp.readline()
                while line:
                    if line[0] == '>':
                        # header name
                        name = line.split()[0][1:]
                        if name == header_name:
                            init_loc = fp.tell()  # file offset of the first NT of the corresponding sequence
                            break

                    line = fp.readline()

                if init_loc is None:
                    print(f"Did not find header name {header_name} in file {file} !!")
                    return ''
                else:
                    fp.seek(init_loc)
                    fp.readline()
                    line_size = fp.tell() - init_loc
                    l_size = line_size - 1
                    self.files_info[token] = {'init_loc': init_loc, 'l_size': l_size, 'line_size': line_size}

            offset_num_lines, offset_in_line = offset // l_size, np.mod(offset, l_size)
            # accounting for extra characters due to multiple \n that will be discarded
            add_for_newline = ((offset + size - 1) // l_size) - offset_num_lines
            fp.seek(init_loc + offset_num_lines * line_size + offset_in_line)
            return fp.read(size + add_for_newline).replace('\n', '')

    def fasta_gen(self, file: str, segment_size: int, init_offset: int = 0, jump_size: int = 0) -> str:
        '''
        Fasta generator.
        Returns an iterator for reading segments of size segment_size NTs, separated by
        jump_size NTs, starting from an offset (0-based) init_offset.

        To read segments consecutively, set jump_size to 0 (which is the default value).
        '''
        try:
            with open(file, 'rt') as fp:
                # handeling the header
                fp.seek(0, os.SEEK_SET)
                if fp.read(1) == ">":
                    fp.readline()
                    init_loc = fp.tell()
                else:
                    init_loc = 0

                # computing number of characters (including \n) per line
                fp.seek(init_loc)
                fp.readline()
                line_size = fp.tell() - init_loc  # number of NTs per line, including the \n
                ext_newlines = segment_size // line_size  # number of \n in a segment_size read with offset 0
                jump_newlines = jump_size // line_size  # same for jump_size

                # starting from init_offset (accounting for \n)
                fp.seek(init_loc + init_offset + (init_offset // line_size))

                while True:
                    segment = fp.read(segment_size + ext_newlines).replace('\n', '')
                    # might need another single read as ext_newlines assumed reading from beginning of the line
                    if len(segment) < segment_size:
                        segment += fp.read(1).replace('\n', '')

                    if segment != '':
                        yield segment.replace('\n', '')
                    else:
                        break  # raise StopIteration

                    # jump to next segment (the jump is excluding \n)
                    if fp.read(jump_size + jump_newlines).count('\n') > jump_newlines:
                        fp.seek(fp.tell() + 1)

        except IOError as err:
            print(f"Error: {file} not found !! ({err})")




import pickle
import json
import re
from pathlib import Path
from bisect import bisect_left

def is_monotonic(A):
    x, y = [], []
    x.extend(A)
    y.extend(A)
    x.sort()
    y.sort(reverse=True)
    if (x == A or y == A):
        return True
    return False



def available_genes():
    from geney import config_setup
    annotation_path = config_setup['MRNA_PATH'] / 'protein_coding'
    return sorted(list(set([m.stem.split('_')[-1] for m in annotation_path.glob('*')])))

def contains(a, x):
    """returns true if sorted sequence `a` contains `x`"""
    i = bisect_left(a, x)
    return i != len(a) and a[i] == x


def unload_json(file_path):
    with open(file_path, 'r') as f:
        data = json.load(f)
    return data


def dump_json(file_path, payload):
    with open(file_path, 'w') as f:
        json.dump(payload, f)
    return None


def unload_pickle(file_path):

    with open(file_path, 'rb') as f:
        data = pickle.load(f)
    return data


def dump_pickle(file_path, payload):
    with open(file_path, 'wb') as f:
        pickle.dump(payload, f)
    return None


def find_files_by_gene_name(gene_name):
    from geney import config_setup
    mrna_path = config_setup['MRNA_PATH'] / 'protein_coding'
    matching_files = [f for f in mrna_path.glob(f'*_{gene_name}.pkl')]
    if len(matching_files) > 1:
        print(f"Multiple files available ({[f.name for f in matching_files]}).")
    elif len(matching_files) == 0:
        raise FileNotFoundError(f"No files available for gene {gene_name}.")

    return matching_files[0]


def reverse_complement(s: str, complement: dict = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A'} ) -> str:
    '''Performs reverse-complement of a sequence. Default is a DNA sequence.'''
    s_rev = s[::-1]
    lower = [b.islower() for b in list(s_rev)]
    bases = [complement.get(base, base) for base in list(s_rev.upper())]
    rev_compl = ''.join([b.lower() if l else b for l, b in zip(lower, bases)])
    return rev_compl






import pandas as pd
import random
from pathlib import Path

class TCGACase:
    def __init__(self, df):
        # Here we get a dataframe of mutations within a gene
        self.df = df
        self.calculate_vaf()
        self.space_variants(spacer_size=50)
        self.case_id = df.case_id.tolist()[0]

    def space_variants(self, spacer_size=100, group_likelihood_threshold=0):
        df = self.df
        if df.empty:
            df['group'] = 0
            return self
        values = sorted(df.Start_Position.unique().tolist())
        # groups = [list(group) for key, group in groupby(values, key=lambda x: (x - values[values.index(x) - 1] >
        # spacer_size) if values.index(x) > 0 else False)] Initialize variables
        groups = []
        current_group = []

        # Iterate through the values
        for i in range(len(values)):
            if i == 0:
                current_group.append(values[i])
            else:
                if values[i] - values[i - 1] <= spacer_size:
                    current_group.append(values[i])
                else:
                    groups.append(current_group)
                    current_group = [values[i]]

        # Append the last group if it's not empty
        if current_group:
            groups.append(current_group)

        df.loc[:, 'group'] = 0
        for i, g in enumerate(groups):
            df.loc[df.Start_Position.isin(g), 'group'] = i
        self.df = df
        return self

    def calculate_vaf(self):
        df = self.df
        df = df[df.t_depth > 0]
        df.loc[:, 'vaf'] = df.apply(lambda row: row.t_alt_count / row.t_depth, axis=1)
        self.df = df
        return self

    def find_overlayed_variants(self):
        df = self.df
        mut_counts = df.mut_id.value_counts()
        mut_counts = mut_counts[mut_counts > 1].index

        small_df = df.groupby('mut_id', as_index=False).agg({
            't_depth': 'sum',
            't_alt_count': 'sum',
            't_ref_count': 'sum',
        })

        df = df.drop_duplicates(subset='mut_id', keep='first')

        small_df = small_df[small_df.t_depth > 0]
        small_df['vaf'] = small_df.t_alt_count / small_df.t_depth

        small_df = small_df.set_index('mut_id')
        df.set_index('mut_id', inplace=True)
        df.update(small_df)
        df.reset_index(inplace=True)
        self.df = df
        return self

    def find_epistasis(self, pth=3, rth=0):
        df = self.df
        if df.empty:
            return None
        # df = df[df.t_alt_count > rth].sort_values('Start_Position', ascending=True)
        df = df[(df.t_alt_count > df.t_ref_count / pth) & (df.t_alt_count >= rth)].sort_values('Start_Position',
                                                                                               ascending=True)

        # display(df[['mut_id', 't_alt_count', 't_ref_count']])

        # Group by the group_key
        grouped = df.groupby('group').agg({
            'mut_id': lambda x: '|'.join(x),
            't_alt_count': 'mean',
            't_ref_count': 'mean',
            'case_id': 'first'
        }).reset_index(drop=True)

        # Drop the group_key column
        return grouped[grouped.mut_id.str.contains('\|')][['mut_id', 't_alt_count', 't_ref_count', 'case_id']]


class TCGAGene:
    def __init__(self, gene, cancer_path=Path('/tamir2/cancer_proj/gdc_db/data/filtered_feb_2021/AllGenes/'),
                 valid_cases=None, extra_cols=[], exclude_filters=None, include_filter=None):
        df = pd.read_csv(cancer_path / gene / 'GeneMutTble.txt',
                         usecols=['Variant_Type', 'FILTER', 'vcf_tumor_gt', 'vcf_normal_gt',
                                  'COSMIC', 't_depth', 't_ref_count', 't_alt_count', 'Proj_name',
                                  'HGVSc', 'Chromosome', 'Start_Position', 'Reference_Allele',
                                  'Tumor_Seq_Allele2', 'case_id', 'Gene_name', 'Variant_Type'] + extra_cols,
                         low_memory=False).sort_values('Start_Position', ascending=True)

        if df.empty:
            self.df = df

        else:
            df = df[df.Variant_Type.isin(['SNP', 'INS', 'DEL'])]
            df = df.astype({'Start_Position': int})

            if include_filter is not None:
                df = df[df.FILTER == include_filter]

            elif exclude_filters is not None:
                for exclude_filter in exclude_filters:
                    df = df[~df.FILTER.str.contains(exclude_filter)]

            if valid_cases is not None:
                df = df[df.case_id.isin(valid_cases)]

            df['mut_id'] = df.apply(lambda
                                        row: f"{row.Gene_name}:{row.Chromosome.replace('chr', '')}:{row.Start_Position}:{row.Reference_Allele}:{row.Tumor_Seq_Allele2}",
                                    axis=1)

            df['ratio'] = df.t_alt_count + df.t_ref_count
            df = df[df.ratio > 0]
            df['ratio'] = df.t_alt_count / df.ratio
            self.df = df

    def affected_cases(self, mut_id=None, read_ratio=0, filters=[]):
        if mut_id is None:
            return self.df.case_id.unique().tolist()
        df = self.df
        df = df[(df.mut_id == mut_id) & (df.ratio >= read_ratio)]
        for filter in filters:
            df = df[~df.FILTER.str.contains(filter)]
        return df.case_id.unique().tolist()

    def get_patient_muts(self, case_id=None, read_ratio=0, exclude_filters=None):
        if case_id is None:
            case_id = random.choice(self.affected_cases())
        return self.df[self.df.case_id == case_id]

    def get_patients_affected(self, mut_id, read_ratio=0, exclude_filters=None):
        # returns all patients affected by ALL the mutatins in mut_id
        pass

    def get_patients_unaffected(self, mut_id, must_contain_all=False, read_ratio=0, exclude_filters=None):
        # returns all patients not affected by ALL the mutation in mut id (patients containg individual mutations only allowed) unless must_contain_all= True
        pass

    def split_patients(self, mut_id, strict=True):
        # returns two lists: all patients affected by a mutation and all patients with none of the mutations (or the mutations but not togehter)
        pass

    def arrange_patients_by_project(self, mut_id):
        # returns all the patients affected by a mutation grouped by cancer project
        pass

    def total_prevalence(self, mut_id):
        pass

    def project_prevalence(self, mut_id):
        pass

    def project_counts(self, mut_id):
        pass









# CLINICAL_DATA_FILE = Path('/tamir2/nicolaslynn/data/TCGA/cancer_reports/new_df_p_proc.pkl')
# CLINICAL_DATA_FILE = Path('/tamir2/yoramzar/Projects/Cancer_mut/Explore_data/reports/df_p_all.pkl')
# CANCER_DATA_PATH = Path('/tamir2/cancer_proj/gdc_db/data/filtered_feb_2021/AllGenes')
# MAF_FILE_NAME = 'GeneMutTble.txt'
# CASE_TRACKER = pd.read_csv('/tamir2/nicolaslynn/projects/TCGAParsed/case2proj.csv', index_col=0)
# PROJ_COUNTS = CASE_TRACKER.proj.value_counts()
# OKGP_DATA_FILE = Path('/tamir2/nicolaslynn/projects/1000GenomesProjMutations/parsed_1kgp_mutations_in_target_genes.csv')
# MUTATION_FREQ_DF = pd.read_csv(OKGP_DATA_FILE, index_col=0)
# PROTEIN_ANNOTATIONS = pd.read_csv('/tamir2/nicolaslynn/data/BioMart/protein_annotations.csv').rename(columns={'Interpro start': 'start', 'Interpro end': 'end', 'Interpro Short Description': 'name'})[['Gene stable ID', 'Transcript stable ID', 'start', 'end', 'name']]
# PROTEIN_ANNOTATIONS['length'] = PROTEIN_ANNOTATIONS.apply(lambda row: abs(row.start - row.end), axis=1)

# def prepare_gene_sets():
#     # gene_annotations_file = Path('/tamir2/nicolaslynn/data/COSMIC/cancer_gene_roles.csv')
#     # GENE_DF = pd.read_csv(gene_annotations_file, index_col=0)
#     # all_oncogenes = GENE_DF[GENE_DF.OG==True].index.tolist()
#     # all_oncogenes = list(set(all_oncogenes))
#     return [], [], []
#
# CLIN_DF = prepare_clinical_data()
# TSGS, ONCOGENES, CANCER_GENES = prepare_gene_sets()
#
#
# def generate_survival_quantitative(affected_df, nonaffected_df):
#     if affected_df.empty or nonaffected_df.empty:
#         return np.nan, np.nan, np.nan
#     results = logrank_test(affected_df['duration'], nonaffected_df['duration'],
#                            event_observed_A=affected_df['event'],
#                            event_observed_B=nonaffected_df['event'])
#     p_value = results.p_value
#     kmf = KaplanMeierFitter()
#     kmf.fit(affected_df['duration'], affected_df['event'], label=f'With Epistasis ({len(affected_df)})')
#     times, surv_probs = kmf.survival_function_.index.values, kmf.survival_function_.values.flatten()
#     auc1 = np.trapz(surv_probs, times)
#     kmf.fit(nonaffected_df['duration'], nonaffected_df['event'], label=f'Without Epistasis ({len(nonaffected_df)})')
#     times, surv_probs = kmf.survival_function_.index.values, kmf.survival_function_.values.flatten()
#     auc2 = np.trapz(surv_probs, times)
#     return p_value, auc1, auc2
#
# def generate_survival_pvalue(affected_df, unaffected_df):
#     results = logrank_test(affected_df['duration'], unaffected_df['duration'],
#                            event_observed_A=affected_df['event'],
#                            event_observed_B=unaffected_df['event'])
#
#     p_value = results.p_value
#     kmf = KaplanMeierFitter()
#     # Fit data
#     kmf.fit(affected_df['duration'], affected_df['event'], label=f'Without Epistasis ({len(affected_df)})')
#     ax = kmf.plot()
#
#     kmf.fit(unaffected_df['duration'], unaffected_df['event'], label=f'With Epistasis ({len(unaffected_df)})')
#     kmf.plot(ax=ax)
#     plt.text(5, 0.95, f'pval: {p_value:.3e}')
#     plt.show()
#     return p_value
#
# def get_project_prevalence(cases_affected):
#     ca = [c for c in cases_affected if c in CASE_TRACKER.index]
#     prevalences = CASE_TRACKER.loc[ca].proj.value_counts() / PROJ_COUNTS
#     prevalences.fillna(0, inplace=True)
#     prevalences = prevalences[[i for i in prevalences.index if 'TCGA' in i]]
#     prevalences.index = [s.replace('TCGA', 'prev') for s in prevalences.index]
#     return prevalences
#
# def get_project_counts(cases_affected):
#     ca = [c for c in cases_affected if c in CASE_TRACKER.index]
#     prevalences = CASE_TRACKER.loc[ca].proj.value_counts()
#     prevalences = prevalences[[i for i in prevalences.index if 'TCGA' in i]]
#     prevalences.index = [s.replace('TCGA_', '') for s in prevalences.index]
#     return prevalences
#
# def get_event_consequence(df):
#     assert df.Transcript_ID.nunique() == 1, 'Too many transcripts to return a single consequenc.'
#     return df.iloc[0].Consequence
#
# def get_dbSNP_id(df):
#     return df.iloc[0].dbSNP_RS
#
# def load_variant_file(gene):
#     df = pd.read_csv(CANCER_DATA_PATH / gene / MAF_FILE_NAME, low_memory=False)
#     df['mut_id'] = df.apply(lambda row: f"{row.Gene_name}:{row.Chromosome.replace('chr', '')}:{row.Start_Position}:{row.Reference_Allele}:{row.Tumor_Seq_Allele2}", axis=1)
#     return df
#
# def find_event_data(event):
#     df = load_variant_file(event.gene)
#     if df.empty:
#         return None
#
#     df = df.query \
#         ('Chromosome == @event.chromosome & Start_Position == @event.start & Reference_Allele == @event.ref & Tumor_Seq_Allele2 == @event.alt')
#
#     if df.empty:
#         return None
#
#     if event.transcript_id is not None:
#         df = df[df.Transcript_ID == event.transcript_id]
#     df['mut_id'] = event.event_id
#     return df
#
#
# class GEvent:
#     def __init__(self, event_id, transcript_id=None):
#         self.gene, self.chromosome, self.start, self.ref, self.alt = event_id.split(':')
#         self.transcript_id = transcript_id
#         self.chromosome = f'chr{self.chromosome}'
#         self.start = int(self.start)
#         self.event_id = event_id
#
#
#
# def get_okgp_mutation_frequency(mut_id):
#     if mut_id in MUTATION_FREQ_DF.index:
#         return MUTATION_FREQ_DF.loc[mut_id].cases_affected
#     else:
#         return 0
#
# def get_df_filter_info(df):
#     filter_artifact_values: list = ["oxog", "bPcr", "bSeq"]
#     MuTect2_filters: list = ['Germline risk', 't_lod_fstar', 'alt_allele_in_normal', 'panel_of_normals', 'clustered_events',
#                              'str_contraction', 'multi_event_alt_allele_in_normal', 'homologous_mapping_event', 'triallelic_site']
#     filter_col_name: str = "FILTER_info"  # column name to add to the dataframe
#     filter_info_list: list = []
#     f_cnr_info = {}
#
#     for j, (prj, df_prj) in enumerate(df.groupby('Proj_name')):
#         filter_vals = list(df_prj['FILTER'])
#         num_pass, num_artifacts, num_mutect2_filters = 0, 0, 0
#         for filter_val in filter_vals:
#             num_pass += ('PASS' in filter_val)
#             num_artifacts += any([x in filter_val for x in filter_artifact_values])
#             num_mutect2_filters += any([x in filter_val for x in MuTect2_filters])
#         num_rest = max(0, (len(filter_vals) - num_pass - num_artifacts - num_mutect2_filters))
#         f_cnr_info[str(prj)[5:]] = (num_pass, num_mutect2_filters, num_artifacts, num_rest)
#     return f_cnr_info
#
# def yoram_mutid(row):
#     return f'{row.Gene_name}:{row.Chromosome}:{row.Consequence}:{row.Start_Position}:{row.Reference_Allele}:{row.Tumor_Seq_Allele2}'
#
#
# def annotate_level_two(mut_id, tid):
#     mut = GEvent(mut_id, tid)
#     df = find_event_data(mut)
#
#     if df.empty or df is None:
#         return None
#
#     patients_affected = df.cases_affected.unique().tolist()
#     p_val, auc_a, auc_n = generate_survival_quantitative(CLIN_DF[CLIN_DF.case_id.isin(patients_affected)], CLIN_DF[~CLIN_DF.case_id.isin(patients_affected)])
#     project_prevalences = get_project_prevalence(patients_affected)
#     prev_dict = project_prevalences.to_dict().sort()
#     project_counts = get_project_counts(patients_affected)
#
#     s = pd.Series({
#         'mut_id': mut_id,
#         'yoram_mut_id': yoram_mutid(df.iloc[0]),
#         'transcript_id': tid,
#         'affected_cases': len(patients_affected),
#         'dbSNP_id': get_dbSNP_id(df),
#         'consequence': get_event_consequence(df),
#         'survival_p_value': p_val,
#         'auc_affected': auc_a,
#         'auc_nonaffected': auc_n,
#         'TSG': contains(TSGS, mut.gene),
#         'oncogene': contains(ONCOGENES, mut.gene),
#         'cases_1kgp': get_okgp_mutation_frequency(mut.event_id),
#         'filter_inf': get_df_filter_info(df),
#         'strand': df.Strand.unique().tolist()[0],
#         'prevalences': prev_dict
#     })
#
#     s['max_prev'] = project_prevalences.max()
#     s['rel_proj'] = ','.join([c.split('_')[-1] for c in project_prevalences[project_prevalences == project_prevalences.max()].index.tolist()])
#     s = pd.concat([s, project_prevalences, project_counts])
#     del df
#     return s
#
# def get_mut_counts():
#     cases = unload_json('/tamir2/nicolaslynn/projects/TCGAParsed/recurring_single_muts_tcga.json')
#     cases = pd.Series(cases)
#     cases.name = 'num_cases'
#     cases.index.name = 'mut_id'
#     cases = cases.to_frame()
#     cases.reset_index(inplace=True)
#     return cases
#
#
# def create_mut_id(row):
#     return f"{row.Gene_name}:{row['Chromosome']}:{row['Start_Position']}:{row['Reference_Allele']}:{row['Tumor_Seq_Allele2']}"
#
#
# def is_in_exon(mut_id, tid):
#     from geney.Gene import Gene
#     transcript = Gene(mut_id.split(':')[0]).generate_transcript(tid)
#     return int(mut_id.split(':')[2]) in transcript.exonic_indices



from .config_setup import get_config
config_setup = get_config()

# import os
# import json
# from pathlib import Path
#
# config_file = os.path.join(os.path.expanduser('~'), '.oncosplice_setup', 'config.json')
# if Path(config_file).exists():
#     config_setup = {k: Path(p) for k, p in json.loads(open(config_file).read()).items()}
#
# else:
#     print("Database not set up.")
#     config_setup = {}
#



import setuptools

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name='geney',
    version='1.1.12',
    python_requires='>3.9',
    description='A Python package for gene expression modeling.',
    url='https://github.com/nicolaslynn/geney',
    author='Nicolas Lynn',
    author_email='nicolasalynn@gmail.com',
    license='Free for non-commercial use',
    packages=setuptools.find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Science/Research',
        'License :: Free for non-commercial use',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3.9',
    ],
)






import os
import json
from pathlib import Path

def get_config():
    config_file = os.path.join(os.path.expanduser('~'), '.oncosplice_setup', 'config.json')
    if Path(config_file).exists():
        config_setup = {k: Path(p) for k, p in json.loads(open(config_file).read()).items()}

    else:
        print("Database not set up.")
        config_setup = {}

    return config_setup


import subprocess
import time
from dask_jobqueue import PBSCluster, SLURMCluster
from dask.distributed import Client, wait
import os
from tqdm import tqdm
from pathlib import Path
from geney import config_setup
from geney.utils import contains, available_genes
import warnings
import gc
import pandas as pd
import argparse

tqdm.pandas()
warnings.filterwarnings('ignore')


def write_executors(folder_path, script='geney.power_utils', input_file='/tamir2/nicolaslynn/data/ClinVar/clinvar_oncosplice_input.txt', output_folder='/tamir2/nicolaslynn/data/oncosplice_base/oncosplice/clinvar_benchmarking', default_logging_path='/tamir2/nicolaslynn/temp/logging'):
    executor_path = Path(folder_path)
    executor_path.mkdir(parents=True, exist_ok=True)
    print(executor_path)
    default_logging_path = Path(default_logging_path)
    job_file_content = f'#!/bin/bash\nhostname\nsource /a/home/cc/students/outside/nicolaslynn/.bashrc\ncd {str(executor_path)}\nsource /tamir2/nicolaslynn/venvs/geney_dask/bin/activate\npython -m {script} -n 750 -m 5GB -i {input_file} -r {output_folder}'
    job_file = executor_path / 'job.sh'
    with open(job_file, 'w') as f:
        _ = f.write(job_file_content)
    submit_file_content = f'qsub -q tamirQ -l nodes=1:ppn=1,cput=24:00:00,mem=25000mb,pmem=25000mb,vmem=50000mb,pvmem=50000mb -e {default_logging_path / "err"} -o {default_logging_path / "out"} {executor_path / "job.sh"}'
    submit_file = executor_path / 'submit'
    with open(submit_file, 'w') as f:
        _ = f.write(submit_file_content)
    subprocess.run(['bash', (executor_path / 'submit')])
    # time.sleep(60)
    # job_file.unlink()
    # submit_file.unlink()
    # executor_path.rmdir()
    return None

def launch_dask_cluster(memory_size="3GB", num_workers=10, queue="tamirQ",
                        walltime="24:00:00", dashboard_address=":23154",
                        log_directory="dask-logs", slurm=False):
    """
    Launch a Dask cluster using PBS.

    Parameters:
    memory_size (str): Memory for each worker.
    num_workers (int): Number of workers to scale to.
    queue (str): Queue name for PBS.
    walltime (str): Walltime for PBS.
    dashboard_address (str): Address for the Dask dashboard.
    log_directory (str): Directory for Dask logs.

    Returns:
    tuple: A tuple containing the Dask client and cluster objects.
    """
    try:
        if slurm:
            dask_cluster = SLURMCluster(
                cores=1,
                memory=memory_size,
                processes=1,
                queue=queue,
                walltime=walltime,
                scheduler_options={"dashboard_address": dashboard_address},
                log_directory=log_directory,
                job_script_prologue=[f"cd {config_setup['BASE']}"]
            )

        else:
            dask_cluster = PBSCluster(
                cores=1,
                memory=memory_size,
                processes=1,
                queue=queue,
                walltime=walltime,
                scheduler_options={"dashboard_address": dashboard_address},
                log_directory=log_directory,
                job_script_prologue=[f"cd {config_setup['BASE']}"]
            )

        dask_cluster.scale(num_workers)
        dask_client = Client(dask_cluster)
        return dask_client, dask_cluster

    except Exception as e:
        print(f"An error occurred: {e}")
        return None, None


def process_and_save_tasks(tasks, dask_client, funct, save_loc=None, num_workers=10, save_increment=20, file_index=0):
    """
    Process a list of tasks using Dask, saving the results incrementally.
    Parameters:
    tasks (list): List of tasks to be processed.
    save_loc (str): Location to save results.
    dask_client (Client): Dask client for task submission.
    num_workers (int): Number of workers to use.
    save_increment (int): Number of iterations after which to save results.
    file_index (int): Starting index for output files.
    Returns:
    None
    """
    def save_results(results, index):
        if results:
            df = pd.concat(results)
            df.to_csv(os.path.join(save_loc, f'results_{index}.csv'))
            return []
        return results

    futures, all_results = [], []
    for i, task in tqdm(enumerate(tasks), total=len(tasks)):
        futures.append(dask_client.submit(funct, task))
        if (i + 1) % num_workers == 0:
            wait(futures)
            all_results.extend([f.result() for f in futures if f.status == 'finished' and f.result() is not None])
            futures = []

        if (i + 1) % (save_increment * num_workers) == 0:
            all_results = save_results(all_results, file_index)
            file_index += 1
            gc.collect()
    wait(futures)
    all_results.extend([f.result() for f in futures if f.status == 'finished' and f.result() is not None])
    save_results(all_results, file_index)


def restart_checkpoint(result_dir):
    """
    Reloads processed results from CSV files, extracting unique mutation IDs and the highest checkpoint.

    Parameters:
    result_dir (str): Directory containing result CSV files.

    Returns:
    list: List of unique mutation IDs processed.
    int: The highest checkpoint value from the files.
    """
    result_path = Path(result_dir)
    files = sorted(result_path.glob('*'), key=lambda x: int(x.stem.split('_')[-1]), reverse=True)

    if not files:
        return [], 0

    try:
        data = []
        latest_file = files[0]
        for file in files:
            data.append(pd.read_csv(file))
        processed_muts = pd.concat(data).mut_id.unique().tolist()
        highest_checkpoint = int(latest_file.stem.split('_')[-1])
        return processed_muts, highest_checkpoint

    except Exception as e:
        print(f"Error processing file {files}: {e}")
        return [], 0


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Run oncosplice with dask.')
    parser.add_argument('--input_file', '-i', required=True, help='input text file')
    parser.add_argument('--results_directory', '-r', required=False, help='result directory', default=config_setup['ONCOSPLICE'])
    parser.add_argument('--num_workers', '-n', type=int, required=False, help='number of dask workers to recruit', default=10)
    parser.add_argument('--worker_size', '-m', type=str, required=False, help='dask worker memory allocation', default="3GB")
    args = parser.parse_args()

    client, cluster = launch_dask_cluster(memory_size=args.worker_size, num_workers=args.num_workers)
    muts = open(args.input_file, 'r').read().splitlines()
    processed_mutations, last_count = restart_checkpoint(args.results_directory)
    processed_mutations = sorted(list(set(processed_mutations)))
    muts = [m for m in tqdm(muts) if not contains(processed_mutations, m)]
    valid_genes = available_genes()
    muts = [m for m in muts if contains(valid_genes, m.split(':')[0])]
    print(f"Valid mutations: {len(muts)}")
    process_and_save_tasks(tasks=muts,
                           save_loc=args.results_directory,
                           dask_client=client,
                           file_index=last_count + 1,
                           num_workers=args.num_workers)
    print("Done.")

